from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred


class UserImportApiABC(metaclass=ABCMeta):
    @abstractmethod
    def importInternalUsers(
        self, importHash: str, usersEncodedPayload: bytes
    ) -> Deferred:
        """Import Internal Users

        Add, replace and remove users in the internal DB

        :param importHash: A string representing this group of items to import
        :param usersEncodedPayload: A List[ImportInternalUserTuple] to import,
                wrapped in a serialised payload.

        Wrap the disps list with ::

                dispsVortexMsg = Payload(tuples=users).toVortexMsg()

        Calling this method with no tuples will delete all items with this importHash


        :return: A deferred that fires when the users are loaded

        """

    @abstractmethod
    def importInternalGroups(
        self, importHash: str, groupsEncodedPayload: bytes
    ) -> Deferred:
        """Import Internal Groups

        Add, replace and remove users in the internal DB

        :param importHash: A string representing this group of items to import
        :param groupsEncodedPayload: A List[ImportInternalGroupTuple] to import,
                wrapped in a serialised payload.

        Wrap the disps list with ::

                dispsVortexMsg = Payload(tuples=groups).toVortexMsg()

        Calling this method with no tuples will delete all items with this importHash


        :return: A deferred that fires when the groups are loaded

        """



from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_core_user._private.PluginNames import userPluginFilt, userPluginObservableName
from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)


def makeTupleDataObservableProxy():
    return TupleDataObservableProxyHandler(
        observableName=userPluginObservableName,
        proxyToVortexName=peekServerName,
        additionalFilt=userPluginFilt,
    )



import logging

from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.handler.TupleDataObservableHandler import (
    TupleDataObservableHandler,
    TuplesProviderABC,
)

from peek_core_user._private.server.tuple_providers.UserLoginUiSettingTupleProvider import (
    UserLoginUiSettingTupleProvider,
)
from peek_core_user._private.tuples.UserLoginUiSettingTuple import (
    UserLoginUiSettingTuple,
)
from peek_plugin_base.PeekVortexUtil import peekBackendNames
from peek_core_user._private.PluginNames import userPluginFilt, userPluginObservableName
from peek_core_user._private.server.tuple_providers.GroupDetailTupleProvider import (
    GroupDetailTupleProvider,
)
from peek_core_user._private.server.tuple_providers.UserLoggedInTupleProvider import (
    UserLoggedInTupleProvider,
)
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from peek_core_user.tuples.GroupDetailTuple import GroupDetailTuple
from peek_core_user.tuples.UserListItemTuple import UserListItemTuple
from .tuple_providers.UserListItemTupleProvider import UserListItemTupleProvider

logger = logging.getLogger(__name__)


class Dummy(TuplesProviderABC):
    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict, *args):
        return Payload(filt, tuples=[]).makePayloadEnvelope().toVortexMsg()


def makeTupleDataObservableHandler(dbSessionCreator, ourApi):
    observable = TupleDataObservableHandler(
        observableName=userPluginObservableName,
        additionalFilt=userPluginFilt,
        acceptOnlyFromVortex=peekBackendNames,
    )

    observable.addTupleProvider(
        GroupDetailTuple.tupleName(), GroupDetailTupleProvider(ourApi)
    )

    observable.addTupleProvider(
        UserListItemTuple.tupleName(),
        UserListItemTupleProvider(dbSessionCreator, ourApi),
    )

    observable.addTupleProvider(
        UserLoggedInTuple.tupleType(),
        UserLoggedInTupleProvider(dbSessionCreator, ourApi),
    )

    observable.addTupleProvider(
        UserLoginUiSettingTuple.tupleType(),
        UserLoginUiSettingTupleProvider(dbSessionCreator),
    )

    return observable



""" 
 *  Copyright Synerty Pty Ltd 2016
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""
import logging

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String

from peek_core_user.tuples.UserLoggedInInfoTuple import UserLoggedInInfoTuple
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class UserLoggedIn(Tuple, DeclarativeBase):
    """UserLoggedIn

    This table stores which users are logged into which devices

    """

    __tupleType__ = userPluginTuplePrefix + "UserLoggedIn"
    __tablename__ = "UserLoggedIn"

    id = Column(Integer, primary_key=True, autoincrement=True)
    loggedInDateTime = Column(DateTime(True), nullable=False)
    userName = Column(String, nullable=False)
    userKey = Column(String, nullable=False)
    userUuid = Column(String, nullable=False)
    deviceToken = Column(String(100), unique=True, nullable=False)
    vehicle = Column(String)
    isFieldLogin = Column(Boolean)

    def toTuple(self):
        return UserLoggedInInfoTuple(
            loggedInDateTime=self.loggedInDateTime,
            userName=self.userName,
            deviceToken=self.deviceToken,
            vehicle=self.vehicle,
            isFieldLogin=self.isFieldLogin,
        )



from abc import ABCMeta
from abc import abstractmethod
from typing import List
from typing import Tuple

from sqlalchemy.orm import Session

from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple


class AuthABC(metaclass=ABCMeta):
    FOR_ADMIN = 1
    FOR_OFFICE = 2
    FOR_FIELD = 3

    @abstractmethod
    def checkPassBlocking(
        self, dbSession: Session, userName: str, password: str, forService: int
    ) -> Tuple[List[str], InternalUserTuple]:
        raise NotImplementedError



"""updated settings

Peek Plugin Database Migration Script

Revision ID: 7e3f668edf0e
Revises: 2209cf338e6d
Create Date: 2019-08-03 20:20:32.348282

"""

# revision identifiers, used by Alembic.
revision = "7e3f668edf0e"
down_revision = "2209cf338e6d"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "UserLoggedIn",
        sa.Column("isFieldLogin", sa.Boolean(), nullable=True),
        schema="core_user",
    )
    # ### end Alembic commands ###

    op.execute(
        """
    DELETE FROM core_user."SettingProperty"
    WHERE "key" = 'Allow Multi Device Logins'
    """
    )

    op.execute(
        """
    UPDATE core_user."SettingProperty"
    SET "key" = 'Field Show Login as List'
    WHERE "key" = 'Show Login as List'
    """
    )

    op.execute(
        """
    UPDATE core_user."SettingProperty"
    SET "key" = 'Field Show Vehicle Input'
    WHERE "key" = 'Show Vehicle Input'
    """
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("UserLoggedIn", "isFieldLogin", schema="core_user")
    # ### end Alembic commands ###



import logging

from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_plugin_base.PeekVortexUtil import peekAdminName
from peek_core_user._private.PluginNames import userPluginFilt, userPluginObservableName
from peek_core_user._private.server.admin_tuple_providers.LoggedInUserStatusTupleProvider import (
    LoggedInUserStatusTupleProvider,
)
from peek_core_user._private.server.tuple_providers.GroupDetailTupleProvider import (
    GroupDetailTupleProvider,
)
from peek_core_user._private.server.tuple_providers.UserListItemTupleProvider import (
    UserListItemTupleProvider,
)
from peek_core_user._private.server.tuple_providers.UserLoggedInTupleProvider import (
    UserLoggedInTupleProvider,
)
from peek_core_user._private.tuples.LoggedInUserStatusTuple import (
    LoggedInUserStatusTuple,
)
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from peek_core_user.tuples.GroupDetailTuple import GroupDetailTuple
from peek_core_user.tuples.UserListItemTuple import UserListItemTuple

logger = logging.getLogger(__name__)


def makeAdminTupleDataObservableHandler(
    dbSessionCreator, deviceApi: DeviceApiABC, ourApi
):
    observable = TupleDataObservableHandler(
        observableName=userPluginObservableName,
        additionalFilt=userPluginFilt,
        acceptOnlyFromVortex=peekAdminName,
    )

    observable.addTupleProvider(
        LoggedInUserStatusTuple.tupleName(),
        LoggedInUserStatusTupleProvider(dbSessionCreator, deviceApi),
    )

    observable.addTupleProvider(
        GroupDetailTuple.tupleName(), GroupDetailTupleProvider(ourApi)
    )

    observable.addTupleProvider(
        UserListItemTuple.tupleName(),
        UserListItemTupleProvider(dbSessionCreator, ourApi),
    )

    observable.addTupleProvider(
        UserLoggedInTuple.tupleType(),
        UserLoggedInTupleProvider(dbSessionCreator, ourApi),
    )

    return observable



import logging

import hashlib
from sqlalchemy.orm.exc import NoResultFound
from twisted.internet.defer import Deferred

from peek_core_user._private.storage.InternalUserPassword import InternalUserPassword
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.tuples.InternalUserUpdatePasswordAction import (
    InternalUserUpdatePasswordAction,
)
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.TupleAction import TupleActionABC
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC

logger = logging.getLogger(__name__)


# This is the CRUD handler
class PasswordUpdateController(TupleActionProcessorDelegateABC):
    def __init__(self, ormSessionCreator):
        self.ormSessionCreator = ormSessionCreator

    @staticmethod
    def hashPass(rawPass: str):
        m = hashlib.sha1()
        m.update(b"peek is a secure thingie")
        m.update(rawPass.encode())
        return m.hexdigest()

    @deferToThreadWrapWithLogger(logger)
    def processTupleAction(self, tupleAction: TupleActionABC) -> Deferred:
        """Process Tuple Action

                The method generates the vortexMsg for the vortex to send.
        s
                :param tupleAction: The C{TupleAction} to process.

        """

        assert isinstance(
            tupleAction, InternalUserUpdatePasswordAction
        ), "not isinstance(tupleAction, InternalUserUpdatePasswordAction)"

        ormSession = self.ormSessionCreator()
        try:
            try:
                password = (
                    ormSession.query(InternalUserPassword)
                    .filter(InternalUserPassword.userId == tupleAction.userId)
                    .one()
                )
            except NoResultFound:
                password = InternalUserPassword()
                password.userId = tupleAction.userId
                ormSession.add(password)

            password.password = self.hashPass(tupleAction.newPassword)
            ormSession.commit()

        finally:
            ormSession.close()



from typing import Callable

from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred

from peek_core_user.server.UserAdminAuthApiABC import UserAdminAuthApiABC
from peek_core_user.server.UserFieldHookApiABC import UserFieldHookApiABC
from peek_core_user.server.UserImportApiABC import UserImportApiABC
from peek_core_user.server.UserInfoApiABC import UserInfoApiABC
from peek_core_user.server.UserLoginApiABC import UserLoginApiABC
from peek_core_user.tuples.login.UserLoginResponseTuple import UserLoginResponseTuple
from peek_core_user.tuples.login.UserLogoutResponseTuple import UserLogoutResponseTuple

UserPostLoginHookCallable = Callable[[UserLoginResponseTuple], Deferred]

UserPostLogoutHookCallable = Callable[[UserLogoutResponseTuple], Deferred]


class UserApiABC(metaclass=ABCMeta):
    @property
    @abstractmethod
    def loginApi(self) -> UserLoginApiABC:
        """Login API

        Returns the API class that handles the user logins

        :return A reference to the UserLoginApiABC class
        """

    @property
    @abstractmethod
    def importApi(self) -> UserImportApiABC:
        """Import API

        Returns the API class that can be used to import the internal user/group objects

        :return A reference to the UserImportApiABC class
        """

    @property
    @abstractmethod
    def fieldHookApi(self) -> UserFieldHookApiABC:
        """Hook API

        Returns the API class that can be used to attach callbacks and observers to the
        login/logout process

        :return A reference to the UserFieldHookApiABC class
        """

    @property
    @abstractmethod
    def infoApi(self) -> UserInfoApiABC:
        """Info API

        Returns the API class that provides information about users

        :return A reference to the UserInfoApiABC class
        """

    @property
    @abstractmethod
    def adminAuth(self) -> UserAdminAuthApiABC:
        """Admin Auth API

        Returns the API class admin authentication for the peek-logic-service

        :return A reference to the UserAdminAuthApiABC class
        """



"""Implemented internal directory

Peek Plugin Database Migration Script

Revision ID: 0498d92479c8
Revises: 3fec57d8da58
Create Date: 2017-10-07 17:33:08.316831

"""

# revision identifiers, used by Alembic.
revision = "0498d92479c8"
down_revision = "3fec57d8da58"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.execute('TRUNCATE TABLE core_user."UserLoggedIn" ')

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "InternalGroup",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("groupName", sa.String(length=50), nullable=False),
        sa.Column("groupTitle", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("groupName"),
        sa.UniqueConstraint("groupTitle"),
        schema="core_user",
    )
    op.create_table(
        "InternalUser",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("userName", sa.String(length=50), nullable=False),
        sa.Column("userTitle", sa.String(length=50), nullable=False),
        sa.Column("userUuid", sa.String(length=50), nullable=False),
        sa.Column("password", sa.String(length=50), nullable=False),
        sa.Column("mobile", sa.String(length=50), nullable=True),
        sa.Column("email", sa.String(length=50), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("userName"),
        sa.UniqueConstraint("userTitle"),
        sa.UniqueConstraint("userUuid"),
        schema="core_user",
    )
    op.create_table(
        "InternalUserGroup",
        sa.Column("userId", sa.Integer(), nullable=False),
        sa.Column("groupId", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["groupId"], ["core_user.InternalGroup.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["userId"], ["core_user.InternalUser.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("userId", "groupId"),
        schema="core_user",
    )
    op.create_index(
        "idx_InternalUserGroup_map",
        "InternalUserGroup",
        ["userId", "groupId"],
        unique=True,
        schema="core_user",
    )
    op.add_column(
        "UserLoggedIn",
        sa.Column("deviceToken", sa.String(length=100), nullable=False),
        schema="core_user",
    )
    op.add_column(
        "UserLoggedIn",
        sa.Column("userName", sa.String(length=50), nullable=True),
        schema="core_user",
    )
    op.drop_constraint(
        "UserLoggedIn_clientDeviceToken_key",
        "UserLoggedIn",
        schema="core_user",
        type_="unique",
    )
    op.drop_constraint(
        "UserLoggedIn_userId_key", "UserLoggedIn", schema="core_user", type_="unique"
    )
    op.create_unique_constraint(None, "UserLoggedIn", ["userName"], schema="core_user")
    op.create_unique_constraint(
        None, "UserLoggedIn", ["deviceToken"], schema="core_user"
    )
    op.drop_column("UserLoggedIn", "userId", schema="core_user")
    op.drop_column("UserLoggedIn", "clientDeviceToken", schema="core_user")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "UserLoggedIn",
        sa.Column(
            "clientDeviceToken",
            sa.VARCHAR(length=100),
            autoincrement=False,
            nullable=False,
        ),
        schema="core_user",
    )
    op.add_column(
        "UserLoggedIn",
        sa.Column("userId", sa.VARCHAR(length=50), autoincrement=False, nullable=True),
        schema="core_user",
    )
    op.drop_constraint(None, "UserLoggedIn", schema="core_user", type_="unique")
    op.drop_constraint(None, "UserLoggedIn", schema="core_user", type_="unique")
    op.create_unique_constraint(
        "UserLoggedIn_userId_key", "UserLoggedIn", ["userId"], schema="core_user"
    )
    op.create_unique_constraint(
        "UserLoggedIn_clientDeviceToken_key",
        "UserLoggedIn",
        ["clientDeviceToken"],
        schema="core_user",
    )
    op.drop_column("UserLoggedIn", "userName", schema="core_user")
    op.drop_column("UserLoggedIn", "deviceToken", schema="core_user")
    op.drop_index(
        "idx_InternalUserGroup_map", table_name="InternalUserGroup", schema="core_user"
    )
    op.drop_table("InternalUserGroup", schema="core_user")
    op.drop_table("InternalUser", schema="core_user")
    op.drop_table("InternalGroup", schema="core_user")
    # ### end Alembic commands ###



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Column
from sqlalchemy import Enum
from sqlalchemy import Integer, String, Index
from sqlalchemy.orm import relationship

from peek_core_user.tuples.UserDetailTuple import UserDetailTuple
from vortex.Tuple import Tuple, addTupleType, TupleField

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.storage.InternalUserGroupTuple import (
    InternalUserGroupTuple,
)

logger = logging.getLogger(__name__)


@addTupleType
class InternalUserTuple(Tuple, DeclarativeBase):
    """Internal

    This table doesn't do anything

    """

    __tupleType__ = userPluginTuplePrefix + "InternalUserTuple"
    __tablename__ = "InternalUser"

    id = Column(Integer, primary_key=True, autoincrement=True)
    userName = Column(String, unique=True, nullable=False)
    userKey = Column(String, unique=True, nullable=True)
    userTitle = Column(String, unique=True, nullable=False)
    userUuid = Column(String, unique=True, nullable=False)
    # `authenticationTarget` is a custom type in postgres, see migration script
    #  43df0e05c728_added_user_import_source.py
    authenticationTarget = Column(String, nullable=False)
    # an arbitrary string of user import source
    importSource = Column(String, nullable=False)

    importHash = Column(String)

    mobile = Column(String)
    email = Column(String)

    groups = relationship(
        InternalGroupTuple, secondary=InternalUserGroupTuple.__table__
    )

    #: This field is ussed for the admin-app to edit the groups
    groupIds = TupleField()

    __table_args__ = (Index("idx_InternalUserTable_importHash", importHash),)

    def toUserDetailTuple(self):
        return UserDetailTuple(
            userName=self.userName,
            userTitle=self.userTitle,
            userUuid=self.userUuid,
            mobile=self.mobile,
            email=self.email,
            groupNames=None,
            data=None,
            authenticationTarget=self.authenticationTarget,
            importSource=self.importSource,
        )



"""Drop InternalUser rows from LDAP for userUuid change

Peek Plugin Database Migration Script

Revision ID: d79da9333e2a
Revises: 43df0e05c728
Create Date: 2022-07-26 15:24:24.408496

"""

# revision identifiers, used by Alembic.
revision = "d79da9333e2a"
down_revision = "43df0e05c728"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    op.drop_index(
        "idx_InternalUserTable_importHash",
        table_name="InternalUser",
        schema="core_user",
    )

    op.execute(
        """
        DELETE FROM core_user."InternalUser"
        WHERE "importSource" = 'LDAP';
        """
    )

    op.create_index(
        "idx_InternalUserTable_importHash",
        "InternalUser",
        ["importHash"],
        unique=False,
        schema="core_user",
    )


def downgrade():
    pass



from abc import ABCMeta
from abc import abstractmethod
from typing import List
from typing import Optional

from peek_core_user.tuples.UserDetailTuple import UserDetailTuple
from twisted.internet.defer import Deferred


class UserInfoApiABC(metaclass=ABCMeta):
    @abstractmethod
    def user(self, userName: str) -> Deferred:
        """Users

        :param userName: The userName of the user to retrieve

        :return: A Deferred, firing with Optional[UserDetailTuple]
        """

    @abstractmethod
    def userBlocking(
        self, userName: str, ormSession=None
    ) -> Optional[UserDetailTuple]:
        """User Details for User ID

        Return an instance of c{UserDetailTuple} for the userName provided.
        :param userName: The username to retrive the details for
        :param ormSession: Specify the ormSession to use, otherwise it may close the
                            current session.

        :return UserDetailTuple
        """

    @abstractmethod
    def userByUserKeyBlocking(
        self, userKey: str, ormSession=None
    ) -> Optional[UserDetailTuple]:
        """User Details for User key

        Return an instance of c{UserDetailTuple} for the userName provided.
        :param userKey: The userkey to retrive the details for
        :param ormSession: Specify the ormSession to use, otherwise it may close the
                            current session.

        :return UserDetailTuple
        """

    @abstractmethod
    def users(
        self,
        likeTitle: Optional[str] = None,
        groupNames: Optional[List[str]] = None,
        isFieldLogin: Optional[bool] = None,
    ) -> Deferred:
        """Users

        :param likeTitle: An optional string to look for in the title of the users
        :param groupNames: An optional list of group names to include users for.

        :return: A Deferred, callbacking with a List[UserDetailTuple]
        """

    @abstractmethod
    def groups(self, likeTitle: Optional[str] = None) -> Deferred:
        """Groups

        :param likeTitle: An optional string to look for in the title of the groups

        :return: A Deferred, callbacking with a List[GroupDetailTuple]
        """

    @abstractmethod
    def peekDeviceTokensForUser(self, userName: str) -> Deferred:  # List[str]:
        """Peek Device Tokens for Logged in User

        Return all the peek device tokens for devices this user is logged in to.

        :return A list of Peek Device Tokens
        """

    @abstractmethod
    def peekUserForDeviceToken(
        self, deviceToken
    ) -> Deferred:  # Optional[UserDetailTuple]:
        """Peek User for Device Token

        Return a user detail tuple for a user logged into a device with deviceToken

        :return UserDetail or None
        """

    @abstractmethod
    def peekTokensWithUserDetails(
        self, isFieldDevice: bool
    ) -> Deferred:  # List[FieldDeviceWithUserDetailsTuple]
        """Peek logged-in field device tokens with or without logged-in user
        details

        :return: A list of Peek Device Tokens logged in or not logged in on
        Field app with user details tuples
        """

    @abstractmethod
    def peekLoggedInDeviceTokens(self, isFieldDevice: bool) -> Deferred:
        # List[
        # str]
        """Peek field device tokens with logged-in status

        :return: A list of string of device tokens
        """

    @abstractmethod
    def userLoggedInInfo(self, userName: str, isFieldDevice: bool) -> Deferred:
        """Peek logged-in field device info by userName

        :param userName: The username / userid of the user, EG C917
        :return: A list of user LoggedIn Info
        """



import logging
from typing import List

from peek_core_user._private.server.auth_connectors.InternalAuth import (
    InternalAuth,
)
from peek_core_user._private.server.auth_connectors.LdapAuth import LdapAuth
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.Setting import (
    globalSetting,
    LDAP_AUTH_ENABLED,
    INTERNAL_AUTH_ENABLED_FOR_ADMIN,
)
from peek_plugin_base.storage.DbConnection import DbSessionCreator
from twisted.cred.error import LoginFailed
from vortex.DeferUtil import deferToThreadWrapWithLogger
from sqlalchemy.orm.exc import NoResultFound

logger = logging.getLogger(__name__)


class AdminAuthController:
    def __init__(self, dbSessionCreator: DbSessionCreator):
        self._dbSessionCreator: DbSessionCreator = dbSessionCreator

    def shutdown(self):
        pass

    @deferToThreadWrapWithLogger(logger)
    def check(self, userName, password) -> List[str]:
        if not password:
            raise LoginFailed("Password is empty")

        ormSession = self._dbSessionCreator()
        try:
            lastException = None

            # TRY INTERNAL IF ITS ENABLED
            try:
                if globalSetting(ormSession, INTERNAL_AUTH_ENABLED_FOR_ADMIN):
                    groupNames, _ = InternalAuth().checkPassBlocking(
                        ormSession, userName, password, InternalAuth.FOR_ADMIN
                    )
                    return groupNames

            except Exception as e:
                lastException = e

            # TRY LDAP IF ITS ENABLED
            try:
                if globalSetting(ormSession, LDAP_AUTH_ENABLED):
                    # TODO Make the client tell us if it's for office or field
                    try:
                        internalUser = (
                            ormSession.query(InternalUserTuple)
                            .filter(InternalUserTuple.userName == userName)
                            .one()
                        )
                        userUuid = internalUser.userUuid
                    except NoResultFound:
                        userUuid = None

                    return LdapAuth().checkPassBlocking(
                        ormSession,
                        userName,
                        password,
                        LdapAuth.FOR_ADMIN,
                        userUuid,
                    )

            except Exception as e:
                lastException = e

            if lastException:
                raise lastException

            raise Exception(
                "No authentication handlers are enabled, enable one in settings"
            )

        finally:
            ormSession.close()



"""Added settings table

Peek Plugin Database Migration Script

Revision ID: 3a9cddaf05e5
Revises: 0498d92479c8
Create Date: 2017-10-14 14:58:46.523021

"""

# revision identifiers, used by Alembic.
revision = "3a9cddaf05e5"
down_revision = "0498d92479c8"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "Setting",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        schema="core_user",
    )
    op.create_table(
        "SettingProperty",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("settingId", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(length=50), nullable=False),
        sa.Column("type", sa.String(length=16), nullable=True),
        sa.Column("int_value", sa.Integer(), nullable=True),
        sa.Column("char_value", sa.String(), nullable=True),
        sa.Column("boolean_value", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(
            ["settingId"],
            ["core_user.Setting.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="core_user",
    )
    op.drop_index(
        "idx_InternalUserGroup_map", table_name="InternalUserGroup", schema="core_user"
    )
    op.create_index(
        "idx_SettingProperty_settingId",
        "SettingProperty",
        ["settingId"],
        unique=False,
        schema="core_user",
    )
    op.create_index(
        "idx_InternalUserGroup_map",
        "InternalUserGroup",
        ["userId", "groupId"],
        unique=True,
        schema="core_user",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "idx_InternalUserGroup_map", table_name="InternalUserGroup", schema="core_user"
    )
    op.drop_index(
        "idx_SettingProperty_settingId",
        table_name="SettingProperty",
        schema="core_user",
    )
    op.drop_table("SettingProperty", schema="core_user")
    op.drop_table("Setting", schema="core_user")
    op.create_index(
        "idx_InternalUserGroup_map",
        "InternalUserGroup",
        ["userId", "groupId"],
        unique=True,
        schema="core_user",
    )
    # ### end Alembic commands ###



import logging

from twisted.internet.defer import inlineCallbacks

from peek_core_user._private.storage.InternalGroupTuple import InternalGroupTuple
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user.tuples.UserListItemTuple import UserListItemTuple
from peek_core_user.tuples.GroupDetailTuple import GroupDetailTuple
from peek_core_user.tuples.UserDetailTuple import UserDetailTuple
from vortex.TupleSelector import TupleSelector

from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

logger = logging.getLogger(__name__)

from peek_core_user._private.worker.tasks.UserImportInternalUserTask import (
    importInternalUsers,
)

from peek_core_user._private.worker.tasks.UserImportInternalGroupTask import (
    importInternalGroups,
)


class ImportController:
    def __init__(self):
        pass

    def shutdown(self):
        pass

    def setTupleObserver(self, tupleDataObserver: TupleDataObservableHandler):
        self._tupleDataObserver = tupleDataObserver

    @inlineCallbacks
    def importInternalUsers(self, importHash: str, usersVortexMsg: bytes):
        #: result: InternalUserImportResultTuple
        result = yield importInternalUsers.delay(importHash, usersVortexMsg)

        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(InternalUserTuple.tupleName(), {})
        )

        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(UserDetailTuple.tupleName(), {})
        )

        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(UserListItemTuple.tupleName(), {})
        )

        return result

    @inlineCallbacks
    def importInternalGroups(self, importHash: str, groupsVortexMsg: bytes):
        #: result: InternalGroupImportResultTuple
        result = yield importInternalGroups.delay(importHash, groupsVortexMsg)

        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(InternalGroupTuple.tupleName(), {})
        )

        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(GroupDetailTuple.tupleName(), {})
        )

        return result



import logging

from rx.subjects import Subject
from twisted.internet.defer import inlineCallbacks

from peek_core_user.server.UserApiABC import (
    UserPostLoginHookCallable,
    UserPostLogoutHookCallable,
)
from peek_core_user.server.UserFieldHookApiABC import UserFieldHookApiABC
from peek_core_user.tuples.login.UserLoginResponseTuple import UserLoginResponseTuple
from peek_core_user.tuples.login.UserLogoutResponseTuple import UserLogoutResponseTuple

logger = logging.getLogger(__name__)


class UserFieldHookApi(UserFieldHookApiABC):
    def __init__(self):

        self._postLoginHooks = []
        self._postLoginSubject = Subject()

        self._postLogoutHooks = []
        self._postLogoutSubject = Subject()

    def shutdown(self):
        pass

    def addPostLoginHook(self, callable: UserPostLoginHookCallable) -> None:
        self._postLoginHooks.append(callable)

    def removePostLoginHook(self, callable: UserPostLoginHookCallable) -> None:
        self._postLoginHooks.remove(callable)

    def postLoginObservable(self) -> Subject:
        return self._postLoginSubject

    def addPostLogoutHook(self, callable: UserPostLogoutHookCallable) -> None:
        self._postLogoutHooks.append(callable)

    def removePostLogoutHook(self, callable: UserPostLogoutHookCallable) -> None:
        self._postLogoutHooks.remove(callable)

    def postLogoffObservable(self) -> Subject:
        return self._postLogoutSubject

    @inlineCallbacks
    def callLogoutHooks(self, response: UserLogoutResponseTuple):
        """
        Returns Deferred[UserLogoutResponseTuple]:
        """

        for callable in self._postLogoutHooks:
            yield callable(response)

        self._postLoginSubject.on_next(response)

    @inlineCallbacks
    def callLoginHooks(self, response: UserLoginResponseTuple):
        """
        Returns Deferred[UserLoginResponseTuple]

        """

        for callable in self._postLoginHooks:
            yield callable(response)

        self._postLoginSubject.on_next(response)



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy.sql.schema import Index, ForeignKey

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase
from vortex.Tuple import Tuple, addTupleType

logger = logging.getLogger(__name__)


@addTupleType
class InternalUserGroupTuple(Tuple, DeclarativeBase):
    """Internal User Group Tuple

    This table stores the relationships between the users and the groups.

    """

    __tupleType__ = userPluginTuplePrefix + "InternalUserGroupTuple"
    __tablename__ = "InternalUserGroup"

    userId = Column(
        Integer,
        ForeignKey("InternalUser.id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )

    groupId = Column(
        Integer,
        ForeignKey("InternalGroup.id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )

    __table_args__ = (Index("idx_InternalUserGroup_map", userId, groupId, unique=True),)



from typing import Type

from peek_plugin_base.agent.PluginAgentEntryHookABC import PluginAgentEntryHookABC
from peek_plugin_base.client.PluginClientEntryHookABC import PluginClientEntryHookABC
from peek_plugin_base.server.PluginLogicEntryHookABC import PluginLogicEntryHookABC
from peek_plugin_base.worker.PluginWorkerEntryHookABC import PluginWorkerEntryHookABC

__version__ = '3.2.6'


def peekFieldEntryHook() -> Type[PluginClientEntryHookABC]:
    from peek_core_user._private.client.PluginClientEntryHook import (
        PluginClientEntryHook,
    )

    return PluginClientEntryHook


def peekOfficeEntryHook() -> Type[PluginClientEntryHookABC]:
    from peek_core_user._private.client.PluginClientEntryHook import (
        PluginClientEntryHook,
    )

    return PluginClientEntryHook


def peekAgentEntryHook() -> Type[PluginAgentEntryHookABC]:
    from ._private.agent.AgentEntryHook import AgentEntryHook

    return AgentEntryHook


def peekLogicEntryHook() -> Type[PluginLogicEntryHookABC]:
    from peek_core_user._private.server.PluginLogicEntryHook import PluginLogicEntryHook

    return PluginLogicEntryHook


def peekWorkerEntryHook() -> Type[PluginWorkerEntryHookABC]:
    from ._private.worker.WorkerEntryHook import WorkerEntryHook

    return WorkerEntryHook



from twisted.internet import defer
from twisted.internet.defer import Deferred, inlineCallbacks

from peek_core_user._private.server.controller.PasswordUpdateController import (
    PasswordUpdateController,
)
from peek_core_user._private.tuples.InternalUserUpdatePasswordAction import (
    InternalUserUpdatePasswordAction,
)
from peek_core_user.server.UserApiABC import UserApiABC
from peek_core_user.tuples.login.UserLoginAction import UserLoginAction
from peek_core_user.tuples.login.UserLogoutAction import UserLogoutAction
from vortex.TupleAction import TupleActionABC
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC


class MainController(TupleActionProcessorDelegateABC):
    def __init__(self, dbSessionCreator, ourApi: UserApiABC):
        self._ourApi = ourApi
        self._passwordUpdateController = PasswordUpdateController(dbSessionCreator)

    def shutdown(self):
        self._ourApi = None

    @inlineCallbacks
    def processTupleAction(self, tupleAction: TupleActionABC) -> Deferred:
        if isinstance(tupleAction, UserLoginAction):
            replyTuples = yield self._ourApi.loginApi.login(tupleAction)
            return replyTuples

        if isinstance(tupleAction, UserLogoutAction):
            replyTuples = yield self._ourApi.loginApi.logout(tupleAction)
            return replyTuples

        if isinstance(tupleAction, InternalUserUpdatePasswordAction):
            replyTuples = yield self._passwordUpdateController.processTupleAction(
                tupleAction
            )
            return replyTuples

        raise NotImplementedError(tupleAction.tupleName())



import logging

from vortex.sqla_orm.OrmCrudHandler import OrmCrudHandler

from peek_core_user._private.PluginNames import userPluginFilt
from peek_core_user._private.storage.LdapSetting import LdapSetting

logger = logging.getLogger(__name__)

# This dict matches the definition in the Admin angular app.
filtKey = {"key": "admin.Edit.LdapSetting"}
filtKey.update(userPluginFilt)


# This is the CRUD hander
class __CrudHandler(OrmCrudHandler):
    pass


# This method creates an instance of the handler class.
def makeLdapSettingeHandler(tupleObservable, dbSessionCreator):
    handler = __CrudHandler(dbSessionCreator, LdapSetting, filtKey, retreiveAll=True)

    return handler



"""Make username lower case

Peek Plugin Database Migration Script

Revision ID: b09fe7b3b31b
Revises: 73d65f042834
Create Date: 2022-01-14 13:11:12.040873

"""

# revision identifiers, used by Alembic.
revision = "b09fe7b3b31b"
down_revision = "73d65f042834"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "InternalUser",
        sa.Column(
            "userKey",
            sa.String(length=50),
            unique=True,
            nullable=True,
        ),
        schema="core_user",
    )
    op.execute(
        """
        UPDATE "core_user"."InternalUser" SET "userKey" = LOWER("userName");
        """
    )
    op.alter_column(
        "InternalUser", "userKey", nullable=False, schema="core_user"
    )

    op.add_column(
        "UserLoggedIn",
        sa.Column(
            "userKey",
            sa.String(length=50),
            unique=False,
            nullable=True,
        ),
        schema="core_user",
    )

    op.add_column(
        "UserLoggedIn",
        sa.Column(
            "userUuid",
            sa.String(length=50),
            unique=False,
            nullable=True,
        ),
        schema="core_user",
    )

    op.execute(
        """
        UPDATE "core_user"."UserLoggedIn" SET "userKey" = LOWER("userName");
        """
    )

    op.execute(
        """
        -- delete users that are no longer in `InternalUser` table
        DELETE FROM "core_user"."UserLoggedIn" uli
        WHERE uli."userName" in (
        SELECT uli."userName" FROM "core_user"."UserLoggedIn" uli
            LEFT OUTER JOIN "core_user"."InternalUser" iu
                ON iu."userName" = uli."userName"
            WHERE uli."userName" IS NOT NULL AND iu."userName" IS NULL
        );  

        -- add `userUuid` to logged in users
        UPDATE "core_user"."UserLoggedIn" uli SET "userUuid" = iu."userUuid"
        FROM "core_user"."InternalUser" iu
        WHERE iu."userName" = uli."userName";
        """
    )
    op.alter_column(
        "UserLoggedIn", "userKey", nullable=False, schema="core_user"
    )
    op.alter_column(
        "UserLoggedIn", "userUuid", nullable=False, schema="core_user"
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("InternalUser", "userKey", schema="core_user")
    op.drop_column("UserLoggedIn", "userKey", schema="core_user")
    op.drop_column("UserLoggedIn", "userUuid", schema="core_user")
    # ### end Alembic commands ###



import logging

from peek_core_user._private.server.auth_connectors.AuthABC import AuthABC
from peek_core_user._private.server.auth_connectors.LdapAuth import LdapAuth
from peek_core_user._private.server.controller.PasswordUpdateController import (
    PasswordUpdateController,
)
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.storage.InternalUserGroupTuple import (
    InternalUserGroupTuple,
)
from peek_core_user._private.storage.InternalUserPassword import (
    InternalUserPassword,
)
from peek_core_user.tuples.constants.UserAuthTargetEnum import (
    UserAuthTargetEnum,
)
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.Setting import (
    globalSetting,
    ADMIN_LOGIN_GROUP,
    OFFICE_LOGIN_GROUP,
    MOBILE_LOGIN_GROUP,
)
from peek_core_user.server.UserDbErrors import (
    UserPasswordNotSetException,
    UserNotFoundException,
)
from twisted.cred.error import LoginFailed

logger = logging.getLogger(__name__)


class InternalAuth(AuthABC):
    def checkPassBlocking(self, dbSession, userName, password, forService):

        authenticatedUser = (
            dbSession.query(InternalUserTuple, InternalUserPassword)
            .join(InternalUserPassword, isouter=True)  # effectively `LEFT JOIN`
            .filter(InternalUserTuple.userName == userName)
            .first()
        )

        # if user not found
        if not authenticatedUser or not authenticatedUser.InternalUserTuple:
            raise UserNotFoundException(userName)

        # if user found but user is created by LDAP
        if (
            authenticatedUser.InternalUserTuple.authenticationTarget
            == UserAuthTargetEnum.LDAP
        ):
            # delegate to LDAPAuth
            return LdapAuth().checkPassBlocking(
                dbSession, userName, password, forService
            )

        if not authenticatedUser.InternalUserPassword:
            raise UserPasswordNotSetException(userName)

        passObj = authenticatedUser.InternalUserPassword
        if passObj.password != PasswordUpdateController.hashPass(password):
            raise LoginFailed(
                "Peek InternalAuth: Username or password is incorrect"
            )

        groups = (
            dbSession.query(InternalGroupTuple)
            .join(InternalUserGroupTuple)
            .filter(InternalUserGroupTuple.userId == passObj.userId)
            .all()
        )

        groupNames = [g.groupName for g in groups]

        if forService == self.FOR_ADMIN:
            adminGroup = globalSetting(dbSession, ADMIN_LOGIN_GROUP)
            if adminGroup not in set(groupNames):
                raise LoginFailed(
                    "Peek InternalAuth: User is not apart of an authorised group"
                )

        elif forService == self.FOR_OFFICE:
            officeGroup = globalSetting(dbSession, OFFICE_LOGIN_GROUP)
            if officeGroup not in set(groupNames):
                raise LoginFailed(
                    "Peek InternalAuth: User is not apart of an authorised group"
                )

        elif forService == self.FOR_FIELD:
            fieldGroup = globalSetting(dbSession, MOBILE_LOGIN_GROUP)
            if fieldGroup not in set(groupNames):
                raise LoginFailed(
                    "Peek InternalAuth: User is not apart of an authorised group"
                )

        else:
            raise Exception(
                "Peek InternalAuth: Unhandled forService type %s" % forService
            )

        return groupNames, authenticatedUser.InternalUserTuple



import logging
from datetime import datetime
from typing import List
from typing import Tuple

import pytz
from peek_core_device.server.DeviceApiABC import DeviceApiABC
from twisted.cred.error import LoginFailed
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_user._private.server.api.UserFieldHookApi import UserFieldHookApi
from peek_core_user._private.server.api.UserInfoApi import UserInfoApi
from peek_core_user._private.server.auth_connectors.InternalAuth import (
    InternalAuth,
)
from peek_core_user._private.server.auth_connectors.LdapAuth import LdapAuth
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.Setting import (
    INTERNAL_AUTH_ENABLED_FOR_FIELD,
)
from peek_core_user._private.storage.Setting import (
    INTERNAL_AUTH_ENABLED_FOR_OFFICE,
)
from peek_core_user._private.storage.Setting import LDAP_AUTH_ENABLED
from peek_core_user._private.storage.Setting import globalSetting
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_core_user._private.tuples.LoggedInUserStatusTuple import (
    LoggedInUserStatusTuple,
)
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from peek_core_user.server.UserDbErrors import (
    UserIsNotLoggedInToThisDeviceError,
)
from peek_core_user.tuples.login.UserLoginAction import UserLoginAction
from peek_core_user.tuples.login.UserLoginResponseTuple import (
    UserLoginResponseTuple,
)
from peek_core_user.tuples.login.UserLogoutAction import UserLogoutAction
from peek_core_user.tuples.login.UserLogoutResponseTuple import (
    UserLogoutResponseTuple,
)
from peek_plugin_base.storage.DbConnection import DbSessionCreator

logger = logging.getLogger(__name__)

USER_ALREADY_LOGGED_ON_KEY = "pl-user.USER_ALREADY_LOGGED_ON"
DEVICE_ALREADY_LOGGED_ON_KEY = "pl-user.DEVICE_ALREADY_LOGGED_ON_KEY"


class _ForceLogout:
    def __init__(self, userUuid, deviceToken):
        self._userUuid = userUuid
        self._deviceToken = deviceToken

    def forceDbLogout(self, ormSession):
        ormSession.query(UserLoggedIn).filter(
            UserLoggedIn.userUuid == self._userUuid
        ).filter(UserLoggedIn.deviceToken == self._deviceToken).delete(
            synchronize_session=False
        )

    def notify(self, clientTupleObservable):
        clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                UserLoggedInTuple.tupleType(),
                selector=dict(deviceToken=self._deviceToken),
            )
        )


class LoginLogoutController:
    def __init__(
        self, deviceApi: DeviceApiABC, dbSessionCreator: DbSessionCreator
    ):
        self._deviceApi: DeviceApiABC = deviceApi
        self._fieldServiceHookApi: UserFieldHookApi = None
        self._infoApi: UserInfoApi = None
        self._dbSessionCreator: DbSessionCreator = dbSessionCreator
        self._clientTupleObservable: TupleDataObservableHandler = None
        self._adminTupleObservable: TupleDataObservableHandler = None

    def setup(
        self,
        clientTupleObservable,
        adminTupleObservable,
        hookApi: UserFieldHookApi,
        infoApi: UserInfoApi,
    ):
        self._clientTupleObservable = clientTupleObservable
        self._adminTupleObservable = adminTupleObservable
        self._fieldServiceHookApi = hookApi
        self._infoApi = infoApi

    def shutdown(self):
        self._clientTupleObservable = None
        self._fieldServiceHookApi = None
        self._infoApi = None

    def _checkPassBlocking(
        self,
        ormSession,
        userName,
        password,
        isFieldService: bool,
        userUuid=None,
    ) -> Tuple[List[str], InternalUserTuple]:
        if not password:
            raise LoginFailed("Password is empty")

        # TODO Make the client tell us if it's for office or field

        lastException = None

        forService = InternalAuth.FOR_OFFICE
        if isFieldService:
            forService = InternalAuth.FOR_FIELD

        # TRY INTERNAL IF ITS ENABLED
        try:
            if forService == InternalAuth.FOR_FIELD and globalSetting(
                ormSession, INTERNAL_AUTH_ENABLED_FOR_FIELD
            ):
                return InternalAuth().checkPassBlocking(
                    ormSession, userName, password, forService
                )

            if forService == InternalAuth.FOR_OFFICE and globalSetting(
                ormSession, INTERNAL_AUTH_ENABLED_FOR_OFFICE
            ):
                return InternalAuth().checkPassBlocking(
                    ormSession, userName, password, forService
                )

        except Exception as e:
            lastException = e

        # TRY LDAP IF ITS ENABLED
        try:
            if globalSetting(ormSession, LDAP_AUTH_ENABLED):
                return LdapAuth().checkPassBlocking(
                    ormSession, userName, password, forService, userUuid
                )

        except Exception as e:
            lastException = e

        if lastException:
            raise lastException

        raise Exception(
            "No authentication handlers are enabled, enable one in settings"
        )

    def _checkGroupBlocking(self, ormSession, groups: List[str]):
        pass

    @deferToThreadWrapWithLogger(logger)
    def _logoutInDb(self, logoutTuple: UserLogoutAction):
        """
        Returns Deferred[UserLogoutResponseTuple]
        """

        session = self._dbSessionCreator()
        try:
            # Check if the user is actually logged into this device.
            qry = (
                session.query(UserLoggedIn)
                .filter(UserLoggedIn.userName == logoutTuple.userName)
                .filter(UserLoggedIn.deviceToken == logoutTuple.deviceToken)
            )

            if qry.count() == 0:
                raise UserIsNotLoggedInToThisDeviceError(logoutTuple.userName)

            session.delete(qry.one())
            session.commit()

        finally:
            session.close()

    @inlineCallbacks
    def logout(self, logoutTuple: UserLogoutAction) -> Deferred:
        """Logout

        :param logoutTuple: The tuple containing the information to process
                                for the logout.

        :return A deferred that fires with List[UserLogoutResponseTuple]
        """

        deviceDescription = yield self._deviceApi.deviceDescription(
            logoutTuple.deviceToken
        )

        response = UserLogoutResponseTuple(
            userName=logoutTuple.userName,
            deviceToken=logoutTuple.deviceToken,
            deviceDescription=deviceDescription,
            acceptedWarningKeys=logoutTuple.acceptedWarningKeys,
            succeeded=True,
        )

        if logoutTuple.isFieldService:
            # Give the hooks a chance to fail the logout
            yield self._fieldServiceHookApi.callLogoutHooks(response)

        # If there are no problems, proceed with the logout.
        try:
            if response.succeeded:
                yield self._logoutInDb(logoutTuple)

        finally:
            # Delay this, otherwise the user gets kicked off before getting
            # the nice success message
            reactor.callLater(
                0.05, self._sendLogoutUpdate, logoutTuple.deviceToken
            )

        self._adminTupleObservable.notifyOfTupleUpdateForTuple(
            LoggedInUserStatusTuple.tupleType()
        )

        return response

    def _sendLogoutUpdate(self, deviceToken: str):
        self._clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                UserLoggedInTuple.tupleType(),
                selector=dict(deviceToken=deviceToken),
            )
        )

    @deferToThreadWrapWithLogger(logger)
    def _loginInDb(self, loginTuple: UserLoginAction):
        """
        Returns Deferred[UserLoginResponseTuple]

        """

        userName = loginTuple.userName
        password = loginTuple.password
        acceptedWarningKeys = set(loginTuple.acceptedWarningKeys)
        deviceToken = loginTuple.deviceToken
        vehicle = loginTuple.vehicleId
        isFieldService = loginTuple.isFieldService
        allowMultipleLogins = isFieldService

        forceLogouter = None

        responseTuple = UserLoginResponseTuple(
            userName=userName,
            userToken="Not implemented",
            succeeded=False,
            acceptedWarningKeys=loginTuple.acceptedWarningKeys,
            vehicleId=loginTuple.vehicleId,
        )

        if not deviceToken:
            raise Exception("peekToken must be supplied")

        thisDeviceDescription = self._deviceApi.deviceDescriptionBlocking(
            deviceToken
        )

        userDetail = self._infoApi.userBlocking(userName)

        # user not found
        if not userDetail and isFieldService:
            responseTuple.setFailed()
            return responseTuple

        # check user group and user password
        ormSession = self._dbSessionCreator()
        try:
            # This will login from the internal user if one already exists or
            # login from LDAP and create a user if an internal does not exist
            groups, _ = self._checkPassBlocking(
                ormSession,
                userName,
                password,
                allowMultipleLogins,
                userDetail.userUuid if userDetail else None,
            )
            self._checkGroupBlocking(ormSession, groups)

            userDetail = self._infoApi.userBlocking(userName)
            if not userDetail:
                responseTuple.setFailed()
                return responseTuple

            userKey = userDetail.userKey
            userUuid = userDetail.userUuid
            loginTuple.userName = userDetail.userName
            responseTuple.userName = userDetail.userName
            responseTuple.userDetail = userDetail

            # Find any current login sessions
            userLoggedIn = (
                ormSession.query(UserLoggedIn)
                .filter(UserLoggedIn.userUuid == userUuid)
                .filter(UserLoggedIn.isFieldLogin == isFieldService)
                .all()
            )
            userLoggedIn = userLoggedIn[0] if userLoggedIn else None

            loggedInElsewhere = (
                ormSession.query(UserLoggedIn)
                .filter(UserLoggedIn.deviceToken != deviceToken)
                .filter(UserLoggedIn.userUuid == userUuid)
                .filter(UserLoggedIn.isFieldLogin == isFieldService)
                .all()
            )

            if allowMultipleLogins and len(loggedInElsewhere) not in (0, 1):
                raise Exception(
                    "Found more than 1 ClientDevice for"
                    + (" token %s" % deviceToken)
                )

            loggedInElsewhere = (
                loggedInElsewhere[0] if loggedInElsewhere else None
            )

            sameDevice = userLoggedIn and loggedInElsewhere is None

            # If the user is logged in, but not to this client device, raise exception
            if allowMultipleLogins and userLoggedIn and not sameDevice:
                if USER_ALREADY_LOGGED_ON_KEY in acceptedWarningKeys:

                    forceLogouter = _ForceLogout(
                        userUuid, loggedInElsewhere.deviceToken
                    )

                    forceLogouter.forceDbLogout(ormSession)

                    userLoggedIn = False

                else:
                    otherDeviceDescription = (
                        self._deviceApi.deviceDescriptionBlocking(
                            loggedInElsewhere.deviceToken
                        )
                    )

                    # This is false if the logged in device has been removed from
                    # enrollment
                    if otherDeviceDescription:
                        responseTuple.setFailed()
                        responseTuple.addWarning(
                            USER_ALREADY_LOGGED_ON_KEY,
                            "User %s is already logged in, on device %s"
                            % (userName, otherDeviceDescription),
                        )

                        return responseTuple

                    # Else, The old device has been deleted,
                    # Just let them login to the same device.

                    forceLogouter = _ForceLogout(
                        loggedInElsewhere.userUuid,
                        loggedInElsewhere.deviceToken,
                    )
                    forceLogouter.forceDbLogout(ormSession)

            # If we're logging into the same device, but already logged in
            if sameDevice:  # Logging into the same device
                sameDeviceDescription = (
                    self._deviceApi.deviceDescriptionBlocking(
                        userLoggedIn.deviceToken
                    )
                )

                responseTuple.deviceToken = userLoggedIn.deviceToken
                responseTuple.deviceDescription = sameDeviceDescription
                responseTuple.succeeded = True
                return responseTuple

            anotherUserOnThatDevice = (
                ormSession.query(UserLoggedIn)
                .filter(UserLoggedIn.deviceToken == deviceToken)
                .filter(UserLoggedIn.userUuid != userUuid)
                .all()
            )

            if anotherUserOnThatDevice:
                anotherUserOnThatDevice = anotherUserOnThatDevice[0]
                if DEVICE_ALREADY_LOGGED_ON_KEY in acceptedWarningKeys:
                    forceLogouter = _ForceLogout(
                        anotherUserOnThatDevice.userUuid,
                        anotherUserOnThatDevice.deviceToken,
                    )
                    forceLogouter.forceDbLogout(ormSession)

                else:
                    responseTuple.setFailed()
                    responseTuple.addWarning(
                        DEVICE_ALREADY_LOGGED_ON_KEY,
                        "User %s is currently logged into this device : %s"
                        % (
                            anotherUserOnThatDevice.userName,
                            thisDeviceDescription,
                        ),
                    )

                    return responseTuple

            # Create the user logged in entry

            newUser = UserLoggedIn(
                userName=userName,
                userKey=userKey,
                userUuid=userUuid,
                loggedInDateTime=datetime.now(pytz.utc),
                deviceToken=deviceToken,
                vehicle=vehicle,
                isFieldLogin=isFieldService,
            )
            ormSession.add(newUser)
            ormSession.commit()

            # Respond with a successful login
            responseTuple.deviceToken = deviceToken
            responseTuple.deviceDescription = thisDeviceDescription
            responseTuple.succeeded = True
            return responseTuple

        finally:
            ormSession.close()

            if forceLogouter:
                forceLogouter.notify(self._clientTupleObservable)

    @inlineCallbacks
    def login(self, loginTuple: UserLoginAction):
        """
        Returns Deferred[UserLoginResponseTuple]

        """
        loginResponse = None
        try:
            loginResponse = yield self._loginInDb(loginTuple)

            if loginTuple.isFieldService:
                yield self._fieldServiceHookApi.callLoginHooks(loginResponse)

        # except UserAlreadyLoggedInError as e:
        #     pass
        #
        # except DeviceAlreadyLoggedInError as e:
        #     pass
        #
        # except UserIsNotLoggedInToThisDeviceError as e:
        #     pass

        except Exception as e:

            # Log the user out again if the hooks fail
            logoutTuple = UserLogoutAction(
                userName=loginTuple.userName, deviceToken=loginTuple.deviceToken
            )

            # Force logout, we don't care if it works or not.
            try:
                yield self._logoutInDb(logoutTuple)
            except UserIsNotLoggedInToThisDeviceError:
                pass

            raise e

        self._clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                UserLoggedInTuple.tupleType(),
                selector=dict(deviceToken=loginTuple.deviceToken),
            )
        )

        self._adminTupleObservable.notifyOfTupleUpdateForTuple(
            LoggedInUserStatusTuple.tupleType()
        )

        return loginResponse

    def _forceLogout(self, ormSession, userUuid, deviceToken):

        ormSession.query(UserLoggedIn).filter(
            UserLoggedIn.userUuid == userUuid
        ).filter(UserLoggedIn.deviceToken == deviceToken).delete(
            synchronize_session=False
        )

        self._clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                UserLoggedInTuple.tupleType(),
                selector=dict(deviceToken=deviceToken),
            )
        )



import logging

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user.tuples.UserListItemTuple import UserListItemTuple
from vortex.Tuple import addTupleType, Tuple, TupleField

logger = logging.getLogger(__name__)


@addTupleType
class UserLoggedInTuple(Tuple):
    """User Logged In Tuple

    This tuple is sent to the devices when a user logs in.

    If the device receives this tuple and the deviceToken doesn't match the current
    device, then the user is logged off.

    """

    __tupleType__ = userPluginTuplePrefix + "UserLoggedInTuple"

    userDetails: UserListItemTuple = TupleField()
    deviceToken: str = TupleField()



import logging

from twisted.internet.defer import Deferred, inlineCallbacks

from peek_core_user._private.server.api.UserApi import UserApi
from peek_core_user._private.storage.Setting import globalSetting, MOBILE_LOGIN_GROUP
from peek_core_user.tuples.UserListItemTuple import UserListItemTuple
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class UserListItemTupleProvider(TuplesProviderABC):
    def __init__(self, dbSessionCreator, ourApi: UserApi):
        self._dbSessionCreator = dbSessionCreator
        self._ourApi = ourApi

        from peek_core_user.server.UserApiABC import UserApiABC

        assert isinstance(self._ourApi, UserApiABC), "We didn't get a UserApiABC"

    @inlineCallbacks
    def makeVortexMsg(self, filt: dict, tupleSelector: TupleSelector) -> Deferred:
        # TODO HACK - Hard coded for mobile users
        mobileGroup = yield self._mobileGroup()
        users = yield self._ourApi.infoApi.users(groupNames=[mobileGroup])

        tuples = []
        for userDetails in users:
            tuples.append(
                UserListItemTuple(
                    userId=userDetails.userName, displayName=userDetails.userTitle
                )
            )

        payload = Payload(filt=filt, tuples=tuples)
        paylodEnvelope = yield payload.makePayloadEnvelopeDefer()
        vortexMsg = yield paylodEnvelope.toVortexMsgDefer()
        return vortexMsg

    @deferToThreadWrapWithLogger(logger)
    def _mobileGroup(self):
        ormSession = self._dbSessionCreator()
        try:
            return globalSetting(ormSession, MOBILE_LOGIN_GROUP)
        finally:
            ormSession.close()



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Column
from sqlalchemy import Integer, String, Index

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase
from vortex.Tuple import Tuple, addTupleType

logger = logging.getLogger(__name__)


@addTupleType
class InternalGroupTuple(Tuple, DeclarativeBase):
    """Group Table

    This table contains the user plugin groups, for the internal directory.

    """

    __tupleType__ = userPluginTuplePrefix + "InternalGroupTuple"
    __tablename__ = "InternalGroup"

    id = Column(Integer, primary_key=True, autoincrement=True)
    groupName = Column(String, unique=True, nullable=False)
    groupTitle = Column(String, unique=True, nullable=False)

    importHash = Column(String)

    __table_args__ = (Index("idx_InternalGroupTable_importHash", importHash),)



import hashlib
import logging
from typing import List
from typing import Tuple

from peek_core_user._private.PluginNames import userPluginTuplePrefix

from peek_core_user._private.server.auth_connectors.AuthABC import AuthABC

from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.LdapSetting import LdapSetting
from peek_core_user._private.storage.Setting import (
    globalSetting,
    LDAP_VERIFY_SSL,
)
from peek_core_user.tuples.constants.UserAuthTargetEnum import (
    UserAuthTargetEnum,
)

from twisted.cred.error import LoginFailed

__author__ = "synerty"

logger = logging.getLogger(__name__)

import ldap

"""
{'objectClass': [b'top', b'person', b'organizationalPerson', b'user'], 'cn': [b'attest'],
 'givenName': [b'attest'],
 'distinguishedName': [b'CN=attest,OU=testou,DC=synad,DC=synerty,DC=com'],
 'instanceType': [b'4'], 'whenCreated': [b'20170505160836.0Z'],
 'whenChanged': [b'20190606130621.0Z'], 'displayName': [b'attest'],
 'uSNCreated': [b'16498'],
 'memberOf': [b'CN=Domain Admins,CN=Users,DC=synad,DC=synerty,DC=com',
              b'CN=Enterprise Admins,CN=Users,DC=synad,DC=synerty,DC=com',
              b'CN=Administrators,CN=Builtin,DC=synad,DC=synerty,DC=com'],
 'uSNChanged': [b'73784'], 'name': [b'attest'],
 'objectGUID': [b'\xee\x1bV\x8dQ\xackE\x82\xd9%_\x18\xadjO'],
 'userAccountControl': [b'66048'], 'badPwdCount': [b'0'], 'codePage': [b'0'],
 'countryCode': [b'0'], 'badPasswordTime': [b'132042996316396717'], 'lastLogoff': [b'0'],
 'lastLogon': [b'132042996806397639'], 'pwdLastSet': [b'132042997225927009'],
 'primaryGroupID': [b'513'], 'objectSid': [
    b'\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00D:3|X\x8f\xc7\x08\xe6\xeaV\xc8Q\x04\x00\x00'],
 'adminCount': [b'1'], 'accountExpires': [b'9223372036854775807'], 'logonCount': [b'36'],
 'sAMAccountName': [b'attest'], 'sAMAccountType': [b'805306368'],
 'userPrincipalName': [b'attest@synad.synerty.com'], 'lockoutTime': [b'0'],
 'objectCategory': [b'CN=Person,CN=Schema,CN=Configuration,DC=synad,DC=synerty,DC=com'],
 'dSCorePropagationData': [b'20190606130621.0Z', b'20190606130016.0Z',
                           b'20170506090346.0Z', b'16010101000000.0Z'],
 'lastLogonTimestamp': [b'132042996806397639']}
"""


class LdapNotEnabledError(Exception):
    pass


class LdapAuth(AuthABC):
    def checkPassBlocking(
        self, dbSession, userName, password, forService, userUuid=None
    ):
        """Login User

        :param forService:
        :param dbSession:
        :param userName: The username of the user.
        :param password: The users secret password.
        :param userUuid: The decoded objectSid of the user from LDAP
        :rtype
        """

        assert forService in (1, 2, 3), "Unhandled for service type"

        ldapSettings: List[LdapSetting] = dbSession.query(LdapSetting).all()

        if not globalSetting(dbSession, LDAP_VERIFY_SSL):
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

        if not ldapSettings:
            raise Exception("LDAPAuth: No LDAP servers configured.")

        firstException = None

        for ldapSetting in ldapSettings:
            if forService == self.FOR_ADMIN:
                if not ldapSetting.adminEnabled:
                    continue

            elif forService == self.FOR_OFFICE:
                if not ldapSetting.desktopEnabled:
                    continue

            elif forService == self.FOR_FIELD:
                if not ldapSetting.mobileEnabled:
                    continue

            else:
                raise Exception(
                    "LDAPAuth: Unhandled forService type %s" % forService
                )

            if (
                "@" in userName
                and userName.split("@")[1] != ldapSetting.ldapDomain
            ):
                continue

            try:
                return self._tryLdap(
                    dbSession, userName, password, ldapSetting, userUuid
                )
            except LoginFailed as e:
                if not firstException:
                    firstException = e

        logger.error("Login failed for %s, %s", userName, str(firstException))

        if firstException:
            raise firstException

        raise LoginFailed("LDAPAuth: No LDAP providers found for this service")

    def _tryLdap(
        self,
        dbSession,
        userName,
        password,
        ldapSetting: LdapSetting,
        userUuid=None,
    ) -> Tuple[List[str], InternalUserTuple]:
        try:

            conn = ldap.initialize(ldapSetting.ldapUri)
            conn.protocol_version = 3
            conn.set_option(ldap.OPT_REFERRALS, 0)

            # make the connection
            conn.simple_bind_s(
                "%s@%s" % (userName.split("@")[0], ldapSetting.ldapDomain),
                password,
            )
            if userUuid:
                ldapFilter = (
                    "(&(objectCategory=person)(objectClass=user)(objectSid=%s))"
                    % userUuid
                )
            else:
                ldapFilter = (
                    "(&(objectCategory=person)(objectClass=user)(sAMAccountName=%s))"
                    % userName.split("@")[0]
                )

            dcParts = ",".join(
                ["DC=%s" % part for part in ldapSetting.ldapDomain.split(".")]
            )

            ldapBases = []
            if ldapSetting.ldapOUFolders:
                ldapBases += self._makeLdapBase(
                    ldapSetting.ldapOUFolders, userName, "OU"
                )
            if ldapSetting.ldapCNFolders:
                ldapBases += self._makeLdapBase(
                    ldapSetting.ldapCNFolders, userName, "CN"
                )

            if not ldapBases:
                raise LoginFailed(
                    "LDAPAuth: LDAP OU and/or CN search paths must be set."
                )

            userDetails = None
            for ldapBase in ldapBases:
                ldapBase = "%s,%s" % (ldapBase, dcParts)

                try:
                    # Example Base : 'CN=atuser1,CN=Users,DC=synad,DC=synerty,DC=com'
                    userDetails = conn.search_st(
                        ldapBase, ldap.SCOPE_SUBTREE, ldapFilter, None, 0, 10
                    )

                    if userDetails:
                        break

                except ldap.NO_SUCH_OBJECT:
                    logger.warning("CN or OU doesn't exist : %s", ldapBase)

        except ldap.NO_SUCH_OBJECT:
            raise LoginFailed(
                "LDAPAuth: An internal error occurred, ask admin to check "
                "Attune logs"
            )

        except ldap.INVALID_CREDENTIALS:
            raise LoginFailed("LDAPAuth: Username or password is incorrect")

        if not userDetails:
            raise LoginFailed(
                "LDAPAuth: User doesn't belong to the correct CN/OUs"
            )

        userDetails = userDetails[0][1]

        distinguishedName = userDetails.get("distinguishedName")[0].decode()
        primaryGroupId = userDetails.get("primaryGroupID")[0].decode()
        objectSid = userDetails.get("objectSid")[0]
        # python-ldap doesn't include key `memberOf` in search result
        #  if the user doesn't belong to any groups.
        memberOfSet = set(userDetails.get("memberOf", []))

        decodedSid = LdapAuth._decodeSid(objectSid)
        primaryGroupSid = (
            "-".join(decodedSid.split("-")[:-1]) + "-" + primaryGroupId
        )

        ldapFilter = "(objectSid=%s)" % primaryGroupSid
        primGroupDetails = conn.search_st(
            dcParts, ldap.SCOPE_SUBTREE, ldapFilter, None, 0, 10
        )
        memberOfSet.add(primGroupDetails[0][1].get("distinguishedName")[0])

        # find all it's groups and groups of those groups
        # The magic number in this filter allows us to fetch the groups of
        # a group.
        ldapFilter = (
            "(&(objectCategory=group)(member:1.2.840.113556.1.4.1941:=%s))"
            % (LdapAuth._escapeParensForLdapFilter(distinguishedName),)
        )
        groupDetails = conn.search_st(
            ",".join(distinguishedName.split(",")[1:]),
            ldap.SCOPE_SUBTREE,
            ldapFilter,
            None,
            0,
            10,
        )

        if groupDetails:
            for group in groupDetails:
                groupMemberOf = group[1].get("memberOf", [])
                memberOfSet.update(groupMemberOf)

        groups = []
        for memberOf in memberOfSet:
            group = memberOf.decode().split(",")[0]
            if "=" in group:
                group = group.split("=")[1]
            groups.append(group)

        userTitle = None
        if userDetails["displayName"]:
            userTitle = userDetails["displayName"][0].decode()

        email = None
        if userDetails["userPrincipalName"]:
            email = userDetails["userPrincipalName"][0].decode()

        if not userUuid:
            userUuid = LdapAuth._decodeSid(objectSid)

        if ldapSetting.ldapGroups:
            ldapGroups = set(
                [s.strip() for s in ldapSetting.ldapGroups.split(",")]
            )

            if not ldapGroups & set(groups):
                raise LoginFailed("User is not a member of an authorised group")

        newInternalUser = self._maybeCreateInternalUserBlocking(
            dbSession,
            userName,
            userTitle,
            userUuid,
            email,
            ldapSetting.ldapTitle,
            objectSid,
        )

        return groups, newInternalUser

    def _decodeSid(sid: [bytes]) -> str:
        strSid = "S-"
        sid = iter(sid)

        # Byte 0 is the revision
        revision = next(sid)
        strSid += "%s" % (revision,)

        # Byte 1 is the count of sub-authorities
        countSubAuths = next(sid)

        # Byte 2-7 (big endian) form the 48-bit authority code
        bytes27 = [next(sid) for _ in range(2, 8)]
        authority = int.from_bytes(bytes27, byteorder="big")
        strSid += "-%s" % (authority,)

        for _ in range(countSubAuths):
            # Each is 4 bytes (32-bits) in little endian
            subAuthBytes = [next(sid) for _ in range(4)]
            subAuth = int.from_bytes(subAuthBytes, byteorder="little")
            strSid += "-%s" % (subAuth,)

        return strSid

    def _escapeParensForLdapFilter(value: str) -> str:
        """Escape parenthesis () in a string

        Escape special characters in a string to be able to use it as a value
        in an LDAP filter. `(` are replaced with \28 and `)` are replaced
        with \29 and so on.

        Reference: https://tools.ietf.org/search/rfc2254#page-5

        :return: Escaped string
        """
        # The \ character must always be escaped first
        value = value.replace("\\", "\\5C")

        value = value.replace("(", "\\28")
        value = value.replace(")", "\\29")
        value = value.replace("*", "\\2A")
        value = value.replace("\0", "\\00")
        return value

    def _maybeCreateInternalUserBlocking(
        self,
        dbSession,
        userName,
        userTitle,
        userUuid,
        email,
        ldapName,
        objectSid,
    ) -> InternalUserTuple:

        internalUser = (
            dbSession.query(InternalUserTuple)
            .filter(InternalUserTuple.userUuid == userUuid)
            .first()
        )

        # do no create, return the existing user
        if internalUser:
            if "@" not in internalUser.userKey:
                internalUser.userKey = (
                    internalUser.userName
                    if "@" in internalUser.userName
                    else (
                        "%s@%s"
                        % (
                            internalUser.userName,
                            internalUser.email.split("@")[1],
                        )
                    )
                )
                dbSession.merge(internalUser)
                dbSession.commit()

            return internalUser

        newInternalUser = InternalUserTuple(
            userName=userName,
            userKey=(
                userName
                if "@" in userName
                else ("%s@%s" % (userName, email.split("@")[1]))
            ),
            userTitle="%s (%s)" % (userTitle, ldapName),
            userUuid=userUuid,
            email=email,
            authenticationTarget=UserAuthTargetEnum.LDAP,
            importSource="LDAP",
            # importHash e.g. 'peek_core_user.LDAPAuth:<md5 hash>'
            importHash=f"{userPluginTuplePrefix}{self.__class__.__name__}:{userUuid}",
        )

        try:
            dbSession.add(newInternalUser)
        except Exception as e:
            logger.info(e)
            raise LoginFailed(
                "Failed to create Internal User. Use the full name <username>@<ldap-domain> to login"
            )
        dbSession.commit()
        return newInternalUser

    def _makeLdapBase(self, ldapFolders, userName, propertyName):
        try:
            ldapBases = []
            for folder in ldapFolders.split(","):
                folder = folder.strip()
                if not folder:
                    continue

                parts = []
                for part in folder.split("/"):
                    part = part.strip()
                    if not part:
                        continue
                    parts.append("%s=%s" % (propertyName, part))

                ldapBases.append(",".join(reversed(parts)))

            return ldapBases

        except Exception as e:
            logger.error(
                "Login failed for %s, failed to parse LDAP %s Folders setting",
                propertyName,
                userName,
            )

            logger.exception(e)

            raise LoginFailed(
                "An internal error occurred, ask admin to check Attune logs"
            )



import logging

from sqlalchemy import Column, Boolean
from sqlalchemy import Integer, String
from vortex.Tuple import Tuple, addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class LdapSetting(Tuple, DeclarativeBase):
    """LdapSetting

    This table stores connetions and settings to LDAP servers

    """

    __tupleType__ = userPluginTuplePrefix + "LdapSettingTuple"
    __tablename__ = "LdapSetting"

    id = Column(Integer, primary_key=True, autoincrement=True)
    ldapTitle = Column(String, nullable=False, unique=True)
    ldapDomain = Column(String, nullable=False)
    ldapUri = Column(String, nullable=False)
    ldapCNFolders = Column(String, nullable=True)
    ldapOUFolders = Column(String, nullable=True)
    ldapGroups = Column(String, nullable=True)

    adminEnabled = Column(Boolean, nullable=False, server_default="0")
    desktopEnabled = Column(Boolean, nullable=False, server_default="0")
    mobileEnabled = Column(Boolean, nullable=False, server_default="0")
